import numpy as np
import pytest
from numpy.typing import NDArray

from relife.data import LifetimeData, load_insulator_string, load_power_transformer
from relife.lifetime_model import (
    AcceleratedFailureTime,
    AgeReplacementModel,
    Exponential,
    Gamma,
    Gompertz,
    LogLogistic,
    ProportionalHazard,
    Weibull,
)

DISTRIBUTION_INSTANCES = [
    Exponential(0.00795203),
    Weibull(3.46597395, 0.01227849),
    Gompertz(0.00865741, 0.06062632),
    Gamma(5.3571091, 0.06622822),
    LogLogistic(3.92614064, 0.0133325),
]

NB_COEF = 2

REGRESSION_INSTANCES = [
    ProportionalHazard(Exponential(0.00795203), coefficients=(np.log(2), np.log(2))),
    ProportionalHazard(Weibull(3.46597395, 0.01227849), coefficients=(np.log(2), np.log(2))),
    ProportionalHazard(Gompertz(0.00865741, 0.06062632), coefficients=(np.log(2), np.log(2))),
    ProportionalHazard(Gamma(5.3571091, 0.06622822), coefficients=(np.log(2), np.log(2))),
    ProportionalHazard(LogLogistic(3.92614064, 0.0133325), coefficients=(np.log(2), np.log(2))),
    AcceleratedFailureTime(Exponential(0.00795203), coefficients=(np.log(2), np.log(2))),
    AcceleratedFailureTime(Weibull(3.46597395, 0.01227849), coefficients=(np.log(2), np.log(2))),
    AcceleratedFailureTime(Gompertz(0.00865741, 0.06062632), coefficients=(np.log(2), np.log(2))),
    AcceleratedFailureTime(Gamma(5.3571091, 0.06622822), coefficients=(np.log(2), np.log(2))),
    AcceleratedFailureTime(LogLogistic(3.92614064, 0.0133325), coefficients=(np.log(2), np.log(2))),
]

########################################################################################################################
# DATA FIXTURES
########################################################################################################################


@pytest.fixture
def power_transformer_data():
    return load_power_transformer()


@pytest.fixture
def insulator_string_data():
    return load_insulator_string()


########################################################################################################################
# LIFETIME DATA FIXTURES
########################################################################################################################


@pytest.fixture
def lifetime_data_1d():
    time = np.array([10, 11, 9, 10, 12, 13, 11], dtype=np.float64)
    event = np.array([1, 0, 1, 0, 0, 0, 1], dtype=np.bool_)
    entry = np.array([0, 0, 3, 5, 3, 1, 9], dtype=np.float64)
    return LifetimeData(time, event=event, entry=entry)


@pytest.fixture
def lifetime_data_2d():
    time = np.array([[1, 2], [0, 4], [5, 5], [7, np.inf], [10, 10], [2, 10], [10, 11]], dtype=np.float64)
    entry = np.array([0, 0, 3, 5, 3, 1, 9], dtype=np.float64)
    departure = np.array([4, np.inf, 7, 10, np.inf, 12, np.inf], dtype=np.float64)
    return LifetimeData(time, entry=entry, departure=departure)


########################################################################################################################
# LIFETIME MODEL FIXTURES
########################################################################################################################


@pytest.fixture(
    params=DISTRIBUTION_INSTANCES,
    ids=["Exponential", "Weibull", "Gompertz", "Gamma", "LogLogistic"],
)
def distribution(request):
    return request.param


@pytest.fixture(
    params=REGRESSION_INSTANCES,
    ids=lambda reg: f"{reg.__class__.__name__}({reg.baseline.__class__.__name__})",
)
def regression(request):
    return request.param


########################################################################################################################
# FROZEN LIFETIME FIXTURES
########################################################################################################################


@pytest.fixture(
    params=REGRESSION_INSTANCES,
    ids=lambda reg: f"Frozen{reg.__class__.__name__}({reg.baseline.__class__.__name__})",
)
def frozen_regression(request):
    covar = np.arange(0.0, 0.6, 0.1).reshape(3, 2)
    return request.param.freeze(covar)


@pytest.fixture(
    params=DISTRIBUTION_INSTANCES,
    ids=lambda distri: f"FrozenAgeReplacementModel({distri.__class__.__name__})",
)
def frozen_ar_distribution(request):
    return AgeReplacementModel(request.param).freeze(request.param.isf(0.75))


@pytest.fixture(
    params=REGRESSION_INSTANCES,
    ids=lambda reg: f"FrozenAgeReplacementModel({reg.__class__.__name__}({reg.baseline.__class__.__name__}))",
)
def frozen_ar_regression(request):
    covar = np.arange(0.0, 0.6, 0.1).reshape(3, 2)
    return AgeReplacementModel(request.param).freeze(request.param.isf(0.75, covar), covar)


########################################################################################################################
# LIFETIME MODEL VARIABLES FIXTURES
########################################################################################################################

M = 3  # nb assets
N = 10  # nb points


@pytest.fixture(
    params=[
        np.float64(1),
        np.ones((1,), dtype=np.float64),
        np.ones((N,), dtype=np.float64),
        np.ones((1, 1), dtype=np.float64),
        np.ones((M, 1), dtype=np.float64),
        np.ones((1, N), dtype=np.float64),
        np.ones((M, 3), dtype=np.float64),
    ],
    ids=lambda time: f"time:{time.shape}",
)
def time(request):
    return request.param


@pytest.fixture(
    params=[
        np.float64(0.5),
        np.ones((1,), dtype=np.float64) * 0.5,
        np.ones((N,), dtype=np.float64) * 0.5,
        np.ones((1, 1), dtype=np.float64) * 0.5,
        np.ones((M, 1), dtype=np.float64) * 0.5,
        np.ones((1, N), dtype=np.float64) * 0.5,
        np.ones((M, N), dtype=np.float64) * 0.5,
    ],
    ids=lambda probability: f"probability:{probability.shape}",
)
def probability(request):
    return request.param


@pytest.fixture(
    params=[
        np.ones((NB_COEF,), dtype=np.float64),
        np.ones((1, NB_COEF), dtype=np.float64),
        np.ones((M, NB_COEF), dtype=np.float64),
    ],
    ids=lambda covar: f"covar:{covar.shape}",
)
def covar(request):
    return request.param


@pytest.fixture(
    params=[
        2.0 * np.ones((), dtype=np.float64),
        2.0 * np.ones((1,), dtype=np.float64),
        2.0 * np.ones((N,), dtype=np.float64),
        2.0 * np.ones((1, N), dtype=np.float64),
        2.0 * np.ones((M, 1), dtype=np.float64),
        2.0 * np.ones((M, N), dtype=np.float64),
    ],
    ids=lambda a: f"a:{a.shape}",
)
def integration_bound_a(request):
    return request.param


@pytest.fixture(
    params=[
        8.0 * np.ones((), dtype=np.float64),
        8.0 * np.ones((1,), dtype=np.float64),
        8.0 * np.ones((N,), dtype=np.float64),
        8.0 * np.ones((1, N), dtype=np.float64),
        8.0 * np.ones((M, 1), dtype=np.float64),
        8.0 * np.ones((M, N), dtype=np.float64),
    ],
    ids=lambda b: f"b:{b.shape}",
)
def integration_bound_b(request):
    return request.param


@pytest.fixture(
    params=[
        np.ones((), dtype=np.float64),
        np.ones((1,), dtype=np.float64),
        np.ones((M,), dtype=np.float64),
        np.ones((M, 1), dtype=np.float64),
    ],
    ids=lambda ar: f"ar:{ar.shape}",
)
def ar(request):
    return request.param


@pytest.fixture(
    params=[
        np.ones((), dtype=np.float64),
        np.ones((1,), dtype=np.float64),
        np.ones((M,), dtype=np.float64),
        np.ones((M, 1), dtype=np.float64),
    ],
    ids=lambda a0: f"a0:{a0.shape}",
)
def a0(request):
    return request.param


@pytest.fixture(params=[1, N, (N,), (M, 1), (M, N)], ids=lambda size: f"size:{size}")
def rvs_size(request):
    return request.param


########################################################################################################################
# ECONOMIC FIXTURES
########################################################################################################################


@pytest.fixture(
    params=[
        np.ones((), dtype=np.float64),
        # np.ones((1,), dtype=np.float64),
        np.ones((M,), dtype=np.float64),
        # np.ones((M, 1), dtype=np.float64),
    ],
    ids=lambda cp: f"cp:{cp.shape}",
)
def cp(request):
    return request.param


# M = 3
CF_RANGE = [5, 10, 20, 100, 1000]


@pytest.fixture(
    params=[
        np.array(CF_RANGE[0], dtype=np.float64),  # ()
        # np.array([CF_RANGE[0]], dtype=np.float64), # (1,)
        np.array(CF_RANGE[:M], dtype=np.float64),  # (M,)
        # np.array(CF_RANGE[:M], dtype=np.float64).reshape(-1, 1), # (M, 1)
    ],
    ids=lambda cf: f"cf:{cf.shape}",
)
def cf(request):
    return request.param


@pytest.fixture(params=[0.0, 0.04], ids=lambda rate: f"discounting_rate:{rate}")
def discounting_rate(request):
    return request.param


########################################################################################################################
# FACTORY FIXTURE TO CONTROL IO SHAPES
########################################################################################################################


@pytest.fixture
def expected_out_shape():
    def _expected_out_shape(**kwargs: NDArray[np.float64]) -> tuple[int, ...]:
        def shape_contrib(**kwargs: NDArray[np.float64]):
            yield ()  # yield at least (), in case kwargs is empty
            for k, v in kwargs.items():
                match k:
                    case "covar" if v.ndim == 2:
                        yield v.shape[0], 1
                    case "covar" if v.ndim < 2:
                        yield ()
                    case "cf" | "cp" | "ar" | "a0" if v.ndim == 2 or v.ndim == 0:
                        yield v.shape
                    case "cf" | "cp" | "ar" | "a0" if v.ndim == 1:
                        yield v.size, 1
                    case "size":
                        if isinstance(v, int):
                            if v == 1:
                                yield ()
                            yield (v,)
                        yield v  # it is tuple
                    case _:
                        yield v.shape

        return np.broadcast_shapes(*tuple(shape_contrib(**kwargs)))

    return _expected_out_shape



import numpy as np
from pytest import approx

from relife.lifetime_model import AgeReplacementModel, LeftTruncatedModel


class TestFrozenRegression:

    def test_rvs(self, regression, rvs_size, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.rvs(rvs_size).shape == expected_out_shape(covar=covar, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True, return_entry=True)
        )

    def test_sf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.sf(time).shape == expected_out_shape(time=time, covar=covar)

    def test_hf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.hf(time).shape == expected_out_shape(time=time, covar=covar)

    def test_chf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.chf(time).shape == expected_out_shape(time=time, covar=covar)

    def test_cdf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.cdf(time).shape == expected_out_shape(time=time, covar=covar)

    def test_pdf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.pdf(time).shape == expected_out_shape(time=time, covar=covar)

    def test_ppf(self, regression, probability, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.ppf(probability).shape == expected_out_shape(probability=probability, covar=covar)

    def test_ichf(self, regression, probability, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.ichf(probability).shape == expected_out_shape(probability=probability, covar=covar)

    def test_isf(self, regression, probability, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.isf(probability).shape == expected_out_shape(probability=probability, covar=covar)
        assert frozen_model.isf(np.full(expected_out_shape(probability=probability, covar=covar), 0.5)) == approx(
            np.broadcast_to(frozen_model.median(), expected_out_shape(probability=probability, covar=covar))
        )

    def test_dhf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.dhf(time).shape == expected_out_shape(time=time, covar=covar)

    def test_jac_sf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.jac_sf(time, asarray=True).shape == (frozen_model.nb_params,) + expected_out_shape(
            time=time, covar=covar
        )

    def test_jac_hf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.jac_hf(time, asarray=True).shape == (frozen_model.nb_params,) + expected_out_shape(
            time=time, covar=covar
        )

    def test_jac_chf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.jac_chf(time, asarray=True).shape == (frozen_model.nb_params,) + expected_out_shape(
            time=time, covar=covar
        )

    def test_jac_cdf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.jac_cdf(time, asarray=True).shape == (frozen_model.nb_params,) + expected_out_shape(
            time=time, covar=covar
        )

    def test_jac_pdf(self, regression, time, covar, expected_out_shape):
        frozen_model = regression.freeze(covar)
        assert frozen_model.jac_pdf(time, asarray=True).shape == (frozen_model.nb_params,) + expected_out_shape(
            time=time, covar=covar
        )


class TestFrozenAgeReplacementDistribution:

    def test_rvs(self, distribution, rvs_size, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)

        assert frozen_model.rvs(rvs_size).shape == expected_out_shape(ar=ar, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(ar=ar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True, return_entry=True)
        )

    def test_sf(self, distribution, time, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.sf(time).shape == expected_out_shape(time=time, ar=ar)

    def test_hf(self, distribution, time, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.hf(time).shape == expected_out_shape(time=time, ar=ar)

    def test_chf(self, distribution, time, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.chf(time).shape == expected_out_shape(time=time, ar=ar)

    def test_cdf(self, distribution, time, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.cdf(time).shape == expected_out_shape(time=time, ar=ar)

    def test_pdf(self, distribution, time, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.pdf(time).shape == expected_out_shape(time=time, ar=ar)

    def test_ppf(self, distribution, time, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.ppf(time).shape == expected_out_shape(time=time, ar=ar)

    def test_ichf(self, distribution, probability, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.ichf(probability).shape == expected_out_shape(probability=probability, ar=ar)

    def test_isf(self, distribution, probability, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        out_shape = expected_out_shape(probability=probability, ar=ar)
        assert frozen_model.isf(probability).shape == out_shape
        assert frozen_model.isf(np.full(out_shape, 0.5)) == approx(np.broadcast_to(frozen_model.median(), out_shape))

    def test_moment(self, distribution, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.moment(1).shape == expected_out_shape(ar=ar)
        assert frozen_model.moment(2).shape == expected_out_shape(ar=ar)

    def test_mean(self, distribution, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.mean().shape == expected_out_shape(ar=ar)

    def test_var(self, distribution, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.var().shape == expected_out_shape(ar=ar)

    def test_median(self, distribution, ar, expected_out_shape):
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        assert frozen_model.median().shape == expected_out_shape(ar=ar)

    def test_ls_integrate(self, distribution, integration_bound_a, integration_bound_b, ar, expected_out_shape):
        np.random.seed(10)
        ar = np.random.uniform(2.0, 8.0, size=ar.shape)
        frozen_model = AgeReplacementModel(distribution).freeze(ar)
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, ar=ar
        )

        integration = frozen_model.ls_integrate(np.ones_like, integration_bound_a, integration_bound_b, deg=100)
        assert integration.shape == out_shape
        assert integration == approx(frozen_model.cdf(integration_bound_b) - frozen_model.cdf(integration_bound_a))
        # integral_0^inf x*dF(x)
        integration = frozen_model.ls_integrate(
            lambda x: x, np.zeros_like(integration_bound_a), np.full_like(integration_bound_b, np.inf), deg=100
        )
        assert integration == approx(np.broadcast_to(frozen_model.mean(), out_shape), rel=1e-3)


class TestFrozenAgeReplacementRegression:

    def test_rvs(self, regression, rvs_size, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.rvs(rvs_size).shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(ar=ar, covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True, return_entry=True)
        )

    def test_sf(self, regression, time, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.sf(time).shape == expected_out_shape(time=time, ar=ar, covar=covar)

    def test_hf(self, regression, time, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.hf(time).shape == expected_out_shape(time=time, ar=ar, covar=covar)

    def test_chf(self, regression, time, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.chf(time).shape == expected_out_shape(time=time, ar=ar, covar=covar)

    def test_cdf(self, regression, time, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.cdf(time).shape == expected_out_shape(time=time, ar=ar, covar=covar)

    def test_pdf(self, regression, time, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.pdf(time).shape == expected_out_shape(time=time, ar=ar, covar=covar)

    def test_ppf(self, regression, time, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.ppf(time).shape == expected_out_shape(time=time, ar=ar, covar=covar)

    def test_ichf(self, regression, probability, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.ichf(probability).shape == expected_out_shape(probability=probability, ar=ar, covar=covar)

    def test_isf(self, regression, probability, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        out_shape = expected_out_shape(probability=probability, ar=ar, covar=covar)
        assert frozen_model.isf(probability).shape == out_shape
        assert frozen_model.isf(np.full(out_shape, 0.5)) == approx(np.broadcast_to(frozen_model.median(), out_shape))

    def test_moment(self, regression, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.moment(1).shape == expected_out_shape(ar=ar, covar=covar)
        assert frozen_model.moment(2).shape == expected_out_shape(ar=ar, covar=covar)

    def test_mean(self, regression, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.mean().shape == expected_out_shape(ar=ar, covar=covar)

    def test_var(self, regression, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.var().shape == expected_out_shape(ar=ar, covar=covar)

    def test_median(self, regression, ar, covar, expected_out_shape):
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        assert frozen_model.median().shape == expected_out_shape(ar=ar, covar=covar)

    def test_ls_integrate(self, regression, integration_bound_a, integration_bound_b, ar, covar, expected_out_shape):
        np.random.seed(10)
        ar = np.random.uniform(2.0, 8.0, size=ar.shape)
        frozen_model = AgeReplacementModel(regression).freeze(ar, covar)
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, ar=ar, covar=covar
        )
        integration = frozen_model.ls_integrate(np.ones_like, integration_bound_a, integration_bound_b, deg=100)
        assert integration.shape == out_shape
        assert integration == approx(frozen_model.cdf(integration_bound_b) - frozen_model.cdf(integration_bound_a))
        # integral_0^inf x*dF(x)
        integration = frozen_model.ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            deg=100,
        )
        assert integration == approx(np.broadcast_to(frozen_model.mean(), out_shape), rel=1e-3)


class TestFrozenLeftTruncatedDistribution:

    def test_rvs(self, distribution, rvs_size, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.rvs(rvs_size).shape == expected_out_shape(a0=a0, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(a0=a0, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True, return_entry=True)
        )

    def test_sf(self, distribution, time, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.sf(time).shape == expected_out_shape(time=time, a0=a0)

    def test_hf(self, distribution, time, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.hf(time).shape == expected_out_shape(time=time, a0=a0)

    def test_chf(self, distribution, time, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.chf(time).shape == expected_out_shape(time=time, a0=a0)

    def test_cdf(self, distribution, time, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.cdf(time).shape == expected_out_shape(time=time, a0=a0)

    def test_pdf(self, distribution, time, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.pdf(time).shape == expected_out_shape(time=time, a0=a0)

    def test_ppf(self, distribution, time, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.ppf(time).shape == expected_out_shape(time=time, a0=a0)

    def test_ichf(self, distribution, probability, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.ichf(probability).shape == expected_out_shape(probability=probability, a0=a0)

    def test_isf(self, distribution, probability, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        out_shape = expected_out_shape(probability=probability, a0=a0)
        assert frozen_model.isf(probability).shape == out_shape
        assert frozen_model.isf(np.full(out_shape, 0.5)) == approx(np.broadcast_to(frozen_model.median(), out_shape))

    def test_moment(self, distribution, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.moment(1).shape == expected_out_shape(a0=a0)
        assert frozen_model.moment(2).shape == expected_out_shape(a0=a0)

    def test_mean(self, distribution, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.mean().shape == expected_out_shape(a0=a0)

    def test_var(self, distribution, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.var().shape == expected_out_shape(a0=a0)

    def test_median(self, distribution, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        assert frozen_model.median().shape == expected_out_shape(a0=a0)

    def test_ls_integrate(self, distribution, integration_bound_a, integration_bound_b, a0, expected_out_shape):
        frozen_model = LeftTruncatedModel(distribution).freeze(a0)
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, a0=a0
        )
        integration = frozen_model.ls_integrate(np.ones_like, integration_bound_a, integration_bound_b, deg=100)
        assert integration.shape == out_shape
        assert integration == approx(frozen_model.cdf(integration_bound_b) - frozen_model.cdf(integration_bound_a))
        # integral_0^inf x*dF(x)
        integration = frozen_model.ls_integrate(
            lambda x: x, np.zeros_like(integration_bound_a), np.full_like(integration_bound_b, np.inf), deg=100
        )
        assert integration == approx(np.broadcast_to(frozen_model.mean(), out_shape), rel=1e-3)


class TestLeftTruncatedRegression:

    # def test_args_nb_assets(self, regression):
    #     frozen_distribution = regression.freeze()
    #     assert frozen_distribution.args_nb_assets == 1

    def test_rvs(self, regression, rvs_size, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.rvs(rvs_size).shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
        assert all(
            arr.shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_entry=True)
        )
        assert all(
            arr.shape == expected_out_shape(a0=a0, covar=covar, size=rvs_size)
            for arr in frozen_model.rvs(rvs_size, return_event=True, return_entry=True)
        )

    def test_sf(self, regression, time, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.sf(time).shape == expected_out_shape(time=time, a0=a0, covar=covar)

    def test_hf(self, regression, time, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.hf(time).shape == expected_out_shape(time=time, a0=a0, covar=covar)

    def test_chf(self, regression, time, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.chf(time).shape == expected_out_shape(time=time, a0=a0, covar=covar)

    def test_cdf(self, regression, time, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.cdf(time).shape == expected_out_shape(time=time, a0=a0, covar=covar)

    def test_pdf(self, regression, time, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.pdf(time).shape == expected_out_shape(time=time, a0=a0, covar=covar)

    def test_ppf(self, regression, time, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.ppf(time).shape == expected_out_shape(time=time, a0=a0, covar=covar)

    def test_ichf(self, regression, probability, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.ichf(probability).shape == expected_out_shape(probability=probability, a0=a0, covar=covar)

    def test_isf(self, regression, probability, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        out_shape = expected_out_shape(probability=probability, a0=a0, covar=covar)
        assert frozen_model.isf(probability).shape == out_shape
        assert frozen_model.isf(np.full(out_shape, 0.5)) == approx(np.broadcast_to(frozen_model.median(), out_shape))

    def test_moment(self, regression, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.moment(1).shape == expected_out_shape(a0=a0, covar=covar)
        assert frozen_model.moment(2).shape == expected_out_shape(a0=a0, covar=covar)

    def test_mean(self, regression, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.mean().shape == expected_out_shape(a0=a0, covar=covar)

    def test_var(self, regression, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.var().shape == expected_out_shape(a0=a0, covar=covar)

    def test_median(self, regression, a0, covar, expected_out_shape):
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        assert frozen_model.median().shape == expected_out_shape(a0=a0, covar=covar)

    def test_ls_integrate(self, regression, integration_bound_a, integration_bound_b, a0, covar, expected_out_shape):
        np.random.seed(10)
        a0 = np.random.uniform(2.0, 8.0, size=a0.shape)
        frozen_model = LeftTruncatedModel(regression).freeze(a0, covar)
        # integral_a^b dF(x)
        out_shape = expected_out_shape(
            integration_bound_a=integration_bound_a, integration_bound_b=integration_bound_b, a0=a0, covar=covar
        )
        integration = frozen_model.ls_integrate(np.ones_like, integration_bound_a, integration_bound_b, deg=100)
        assert integration.shape == out_shape
        assert integration == approx(frozen_model.cdf(integration_bound_b) - frozen_model.cdf(integration_bound_a))
        # integral_0^inf x*dF(x)
        integration = frozen_model.ls_integrate(
            lambda x: x,
            np.zeros_like(integration_bound_a),
            np.full_like(integration_bound_b, np.inf),
            deg=100,
        )
        assert integration == approx(np.broadcast_to(frozen_model.mean(), out_shape), rel=1e-3)



from typing import Generic, Optional, Self, TypeVarTuple

import numpy as np
from numpy.typing import NDArray as NDArray

_Xs = TypeVarTuple("_Xs")

class ParametricModel:
    _params: Parameters

    def __init__(self, **kwparams: Optional[float]) -> None: ...
    @property
    def params(self) -> NDArray[np.float64]: ...

    # noinspection PyUnresolvedReferences
    @params.setter
    def params(self, new_params: NDArray[np.float64]): ...
    @property
    def params_names(self) -> tuple[str, ...]: ...
    @property
    def nb_params(self) -> int: ...

class Parameters:
    _parent: Optional[Self]
    _leaves = dict[str, Self]
    _nodemapping = dict[str, float]
    _values: tuple[float, ...]
    _names: tuple[str, ...]

    def __init__(self, **kwargs: Optional[float]) -> None: ...
    def get_names(self) -> tuple[str, ...]: ...
    def get_values(self) -> tuple[float, ...]: ...
    def get_leaf(self, name: str) -> Parameters: ...
    def get_param_value(self, name: str) -> float: ...
    def set_param_value(self, name: str, value: float) -> None: ...
    @property
    def size(self) -> int: ...
    def set_node(self, mapping: dict[str, float | None]) -> None: ...
    def set_leaf(self, leaf_name: str, leaf: Parameters) -> None: ...
    def set_values(self, values: tuple[float, ...]) -> None: ...
    def _update_names_and_values(self) -> None: ...

class FrozenParametricModel(ParametricModel, Generic[*_Xs]):
    nb_assets: int
    args: tuple[*_Xs]
    unfrozen_model: ParametricModel
    def __init__(self, model: ParametricModel, *args: *_Xs) -> None: ...
    def unfreeze(self) -> ParametricModel: ...

def is_frozen(model: ParametricModel) -> bool: ...
def is_lifetime_model(model: ParametricModel) -> bool: ...
def is_stochastic_process(model: ParametricModel) -> bool: ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Iterable,
    Optional,
    TypeVarTuple,
)

import numpy as np
from numpy.typing import NDArray
from typing_extensions import override

from relife.lifetime_model import (
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
)
from relife.stochastic_process import FrozenNonHomogeneousPoissonProcess

from ...lifetime_model.distribution import LifetimeDistribution
from ...lifetime_model.regression import FrozenLifetimeRegression
from .iterators import (
    CountDataIterator,
    NonHomogeneousPoissonProcessIterator,
    RenewalProcessIterator,
    RenewalRewardProcessIterator,
)

if TYPE_CHECKING:
    from relife.stochastic_process import RenewalProcess

Args = TypeVarTuple("Args")


class CountDataIterable(Iterable[NDArray[np.void]], ABC):
    def __init__(
        self,
        size: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        self.size = size
        self.tf = tf
        self.t0 = t0
        self.nb_assets = nb_assets
        self.seed = seed

    @abstractmethod
    @override
    def __iter__(self) -> CountDataIterator: ...


def age_of_renewal_process_sampler(
    lifetime_model,
    nb_samples: int,
    t: float,
    nb_assets: int = 1,
    first_lifetime_model: Optional = None,
    seed: Optional[int] = None,
):
    timeline = np.zeros((nb_assets, nb_samples), dtype=np.float64)
    just_crossed_t = np.zeros_like(timeline, dtype=np.uint32)
    age_process = np.zeros_like(timeline, dtype=np.float64)
    replacement_cycle = 0

    while np.any(timeline < t):
        if replacement_cycle == 0 and first_lifetime_model is not None:
            time, entry = first_lifetime_model.rvs((nb_assets, nb_samples), return_entry=True, seed=seed)
        else:
            time, entry = lifetime_model.rvs((nb_assets, nb_samples), return_entry=True, seed=seed)
        replacement_cycle += 1
        residual_time = time - entry
        timeline += residual_time
        just_crossed_t[timeline > t] += 1
        age_process = np.where(just_crossed_t == 1, time - (timeline - t), age_process)
        if seed is not None:
            seed += 1
    return np.squeeze(age_process)


class RenewalProcessIterable(CountDataIterable):

    def __init__(
        self,
        process: RenewalProcess[
            LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel
        ],
        size: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        super().__init__(size, tf, t0=t0, nb_assets=nb_assets, seed=seed)
        # TODO : control and broadcast size here !
        self.process = process
        if nb_assets is None:
            self.nb_assets = getattr(process.lifetime_model, "nb_assets", 1)

    def __iter__(self) -> RenewalProcessIterator:
        from relife.stochastic_process import RenewalProcess

        if isinstance(self.process, RenewalProcess):
            return RenewalProcessIterator(
                self.process, self.size, self.tf, t0=self.t0, nb_assets=self.nb_assets, seed=self.seed
            )
        else:
            return RenewalRewardProcessIterator(
                self.process, self.size, self.tf, t0=self.t0, nb_assets=self.nb_assets, seed=self.seed
            )


class NonHomogeneousPoissonProcessIterable(CountDataIterable):
    def __init__(
        self,
        process: FrozenNonHomogeneousPoissonProcess,
        size: int,
        tf: float,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        super().__init__(size, tf, t0=t0, nb_assets=nb_assets, seed=seed)
        # TODO : control and broadcast size here !
        self.process = process
        if nb_assets is None:
            self.nb_assets = getattr(process, "nb_assets", 1)

    def __iter__(self) -> NonHomogeneousPoissonProcessIterator:
        return NonHomogeneousPoissonProcessIterator(
            self.process, self.size, self.tf, t0=self.t0, nb_assets=self.nb_assets, seed=self.seed
        )



from .data import (
    NonHomogeneousPoissonProcessSample,
    RenewalProcessSample,
    RenewalRewardProcessSample,
)
from .iterables import NonHomogeneousPoissonProcessIterable, RenewalProcessIterable



"""
This module defines probability functions used in regression

Copyright (c) 2022, RTE (https://www.rte-france.com)
See AUTHORS.txt
SPDX-License-Identifier: Apache-2.0 (see LICENSE.txt)
"""

from abc import ABC

import numpy as np
from scipy.optimize import Bounds

from relife import ParametricModel

from ._base import FittableParametricLifetimeModel, FrozenParametricLifetimeModel


def broadcast_time_covar(time, covar):
    time = np.atleast_2d(np.asarray(time))  #  (m, n)
    covar = np.atleast_2d(np.asarray(covar))  #  (m, nb_coef)
    match (time.shape[0], covar.shape[0]):
        case (1, _):
            time = np.repeat(time, covar.shape[0], axis=0)
        case (_, 1):
            covar = np.repeat(covar, time.shape[0], axis=0)
        case (m1, m2) if m1 != m2:
            raise ValueError(f"Incompatible time and covar. time has {m1} nb_assets but covar has {m2} nb_assets")
    return time, covar


def broadcast_time_covar_shapes(time_shape, covar_shape):
    # time_shape : (), (n,) or (m, n)
    # covar_shape : (), (nb_coef,) or (m, nb_coef)
    match [time_shape, covar_shape]:
        case [(), ()] | [(), (_,)]:
            return ()
        case [(), (m, _)]:
            return m, 1
        case [(n,), ()] | [(n,), (_,)]:
            return (n,)
        case [(n,), (m, _)] | [(m, n), ()] | [(m, n), (_,)]:
            return m, n
        case [(mt, n), (mc, _)] if mt != mc:
            if mt != 1 and mc != 1:
                raise ValueError(f"Invalid time and covar : time got {mt} nb assets but covar got {mc} nb assets")
            return max(mt, mc), n
        case [(mt, n), (mc, _)] if mt == mc:
            return mt, n
        case _:
            raise ValueError(f"Invalid time or covar shape. Got {time_shape} and {covar_shape}")


class CovarEffect(ParametricModel):
    """
    Covariates effect.

    Parameters
    ----------
    *coefficients : float
        Coefficients of the covariates effect.
    """

    def __init__(self, coefficients=(None,)):
        super().__init__(**{f"coef_{i + 1}": v for i, v in enumerate(coefficients)})

    @property
    def nb_coef(self):
        """
        The number of coefficients

        Returns
        -------
        int
        """
        return self.nb_params

    def g(self, covar):
        """
        Compute the covariates effect.
        If covar.shape : () or (nb_coef,) => out.shape : (), float
        If covar.shape : (m, nb_coef) => out.shape : (m, 1)
        """
        arr_covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        if arr_covar.ndim > 2:
            raise ValueError(f"Invalid covar shape. Expected (nb_coef,) or (m, nb_coef) but got {arr_covar.shape}")
        covar_nb_coef = arr_covar.size if arr_covar.ndim <= 1 else arr_covar.shape[-1]
        if covar_nb_coef != self.nb_coef:
            raise ValueError(
                f"Invalid covar. Number of covar does not match number of coefficients. Got {self.nb_coef} nb_coef but covar shape is {arr_covar.shape}"
            )
        g = np.exp(np.sum(self.params * arr_covar, axis=-1, keepdims=True))  # (m, 1)
        if arr_covar.ndim <= 1:
            return np.float64(g.item())
        return g

    def jac_g(self, covar, *, asarray=False):
        """
        Compute the Jacobian of the covariates effect.
        If covar.shape : () or (nb_coef,) => out.shape : (nb_coef,)
        If covar.shape : (m, nb_coef) => out.shape : (nb_coef, m, 1)
        """
        arr_covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        g = self.g(arr_covar)  # () or (m, 1)
        jac = arr_covar.T.reshape(self.nb_coef, -1, 1) * g  # (nb_coef, m, 1)
        if arr_covar.ndim <= 1:
            jac = jac.reshape(self.nb_coef)  # (nb_coef,) or (nb_coef, m, 1)
        if not asarray:
            return np.unstack(jac, axis=0)  # tuple
        return jac  # (nb_coef, m, 1)


class LifetimeRegression(FittableParametricLifetimeModel, ABC):
    """
    Base class for regression model.

    At least one positional covar arg and 0 or more additional args (variable number)
    see : https://peps.python.org/pep-0646/#unpacking-unbounded-tuple-types*

    Note:
    LifetimeRegression does not preserve generic : at the moment, additional args are supposed to be always float | NDArray[np.float64]
    """

    def __init__(self, baseline, coefficients=(None,)):
        super().__init__()
        self.covar_effect = CovarEffect(coefficients)
        self.baseline = baseline
        self.fitting_results = None

    @property
    def coefficients(self):
        """Get the coefficients values of the covariate effect.

        Returns
        -------
        ndarray
        """
        return self.covar_effect.params

    @property
    def nb_coef(self):
        """The number of coefficients.

        Returns
        -------
        int
        """
        return self.covar_effect.nb_params

    def sf(self, time, covar, *args):
        """
        The survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().sf(time, covar, *args)

    def isf(self, probability, covar, *args):
        """
        The inverse survival function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        return self.ichf(cumulative_hazard_rate, covar, *args)

    def cdf(self, time, covar, *args):
        """
        The cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().cdf(time, *(covar, *args))

    def pdf(self, time, covar, *args):
        """
        The probility density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().pdf(time, *(covar, *args))

    def ppf(self, probability, covar, *args):
        """
        The percent point function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        return super().ppf(probability, *(covar, *args))

    def mrl(self, time, covar, *args):
        """
        The mean residual life.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return super().mrl(time, *(covar, *args))

    def ls_integrate(self, func, a, b, covar, *args, deg: int = 10):
        """
        Lebesgue-Stieltjes integration.

        Parameters
        ----------
        func : callable (in : 1 ndarray , out : 1 ndarray)
            The callable must have only one ndarray object as argument and one ndarray object as output
        a : ndarray (maximum number of dimension is 2)
            Lower bound(s) of integration.
        b : ndarray (maximum number of dimension is 2)
            Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.
        deg : int, default 10
            Degree of the polynomials interpolation

        Returns
        -------
        np.ndarray
            Lebesgue-Stieltjes integral of func from `a` to `b`.
        """
        return super().ls_integrate(func, a, b, *(covar, *args), deg=deg)

    def moment(self, n, covar, *args):
        """
        n-th order moment

        Parameters
        ----------
        n : int
            order of the moment, at least 1.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return super().moment(n, *(covar, *args))

    def mean(self, covar, *args):
        """
        The mean.

        Parameters
        ----------
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return super().mean(*(covar, *args))

    def var(self, covar, *args):
        """
        The variance.

        Parameters
        ----------
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return super().var(*(covar, *args))

    def median(self, covar, *args):
        """
        The median.

        Parameters
        ----------
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return super().median(*(covar, *args))

    def jac_sf(
        self,
        time,
        covar,
        *args,
        asarray=False,
    ):
        """
        The jacobian of the survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = -self.jac_chf(time, covar, *args, asarray=True) * self.sf(time, covar, *args)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_cdf(self, time, covar, *args, asarray=False):
        """
        The jacobian of the cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = -self.jac_sf(time, covar, *args, asarray=True)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_pdf(self, time, covar, *args, asarray=False):
        """
        The jacobian of the probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        jac = self.jac_hf(time, covar, *args, asarray=True) * self.sf(time, covar, *args) + self.jac_sf(
            time, covar, *args, asarray=True
        ) * self.hf(time, covar, *args)
        if not asarray:
            return np.unstack(jac)
        return jac

    def rvs(self, size: int, covar, *args, nb_assets=None, return_event=False, return_entry=False, seed=None):
        """
        Random variable sampling.

        Parameters
        ----------
        size : int
            Size of the generated sample.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        nb_assets : int, optional
            If nb_assets is not None, 2d arrays of samples are generated.
        return_event : bool, default is False
            If True, returns event indicators along with the sample time values.
        return_entry : bool, default is False
            If True, returns corresponding entry values of the sample time values.
        seed : optional int, default is None
            Random seed used to fix random sampling.

        Returns
        -------
        float, ndarray or tuple of float or ndarray
            The sample values. If either ``return_event`` or ``random_entry`` is True, returns a tuple containing
            the time values followed by event values, entry values or both.
        """
        return super().rvs(
            size, *(covar, *args), nb_assets=nb_assets, return_event=return_event, return_entry=return_entry, seed=seed
        )

    def _get_initial_params(self, time, covar, *args, event=None, entry=None, departure=None):
        self.covar_effect = CovarEffect(
            (None,) * covar.shape[-1]
        )  # changes params structure depending on number of covar
        param0 = np.zeros_like(self.params, dtype=np.float64)
        param0[-self.baseline.params.size :] = self.baseline._get_initial_params(
            time, *args, event=None, entry=None, departure=None
        )  # recursion in case of PPH(AFT(...))
        return param0

    def _get_params_bounds(self):
        lb = np.concatenate(
            (
                np.full(self.covar_effect.nb_params, -np.inf),
                self.baseline._get_params_bounds().lb,  # baseline has _params_bounds according to typing
            )
        )
        ub = np.concatenate(
            (
                np.full(self.covar_effect.nb_params, np.inf),
                self.baseline._get_params_bounds().ub,
            )
        )
        return Bounds(lb, ub)

    def fit(
        self,
        time,
        covar,
        *args,
        event=None,
        entry=None,
        departure=None,
        **options,
    ):
        """
        Estimation of the regression parameters from lifetime data.

        Parameters
        ----------
        time : ndarray (1d or 2d)
            Observed lifetime values.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        departure : ndarray of float (1d), default is None
            Right truncations applied to lifetime values.
        **options
            Extra arguments used by `scipy.minimize`. Default values are:
                - `method` : `"L-BFGS-B"`
                - `contraints` : `()`
                - `tol` : `None`
                - `callback` : `None`
                - `options` : `None`
                - `bounds` : `self.params_bounds`
                - `x0` : `self.init_params`

        Returns
        -------
        Self
            The current object with the estimated parameters setted inplace.

        Notes
        -----
        Supported lifetime observations format is either 1d-array or 2d-array. 2d-array is more advanced
        format that allows to pass other information as left-censored or interval-censored values. In this case,
        `event` is not needed as 2d-array encodes right-censored values by itself.
        """
        return super().fit(time, *(covar, *args), event=event, entry=entry, departure=departure, **options)

    def freeze(self, covar, *args):
        """
        Freeze regression covar and other arguments into the object data.

        Parameters
        ----------
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenLifetimeRegression
        """
        return FrozenLifetimeRegression(self, covar, *args)


class ProportionalHazard(LifetimeRegression):
    # noinspection PyUnresolvedReferences
    r"""
    Proportional Hazard regression.

    The cumulative hazard function :math:`H` is linked to the multiplier
    function :math:`g` by the relation:

    .. math::

        H(t, x) = g(\beta, x) H_0(t) = e^{\beta \cdot x} H_0(t)

    where :math:`x` is a vector of covariates, :math:`\beta` is the coefficient
    vector of the effect of covariates, :math:`H_0` is the baseline cumulative
    hazard function [1]_.

    |

    Parameters
    ----------
    baseline : FittableParametricLifetimeModel
        Any lifetime model that can be fitted.
    coefficients : tuple of floats (values can be None), default is (None,)
        Coefficients values of the covariate effects.

    Attributes
    ----------
    baseline : FittableParametricLifetimeModel
        The regression baseline model (lifetime model).
    covar_effect : CovarEffect
        The regression covariate effect.
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    coefficients
    nb_params
    params
    params_names
    plot


    References
    ----------
    .. [1] Sun, J. (2006). The statistical analysis of interval-censored failure
        time data (Vol. 3, No. 1). New York: springer.

    See Also
    --------
    regression.AFT : Accelerated Failure Time regression.

    """

    def hf(self, time, covar, *args):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.covar_effect.g(covar) * self.baseline.hf(time, *args)

    def chf(self, time, covar, *args):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.covar_effect.g(covar) * self.baseline.chf(time, *args)

    def ichf(self, cumulative_hazard_rate, covar, *args):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given cumulative hazard rate(s).
        """
        return self.baseline.ichf(cumulative_hazard_rate / self.covar_effect.g(covar), *args)

    def dhf(self, time, covar, *args):
        """
        The derivative of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.covar_effect.g(covar) * self.baseline.dhf(time, *args)

    def jac_hf(self, time, covar, *args, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)

        baseline_hf = self.baseline.hf(time, *args)  # (m, n)
        # p == baseline.nb_params
        baseline_jac_hf = self.baseline.jac_hf(time, *args, asarray=True)  # (p, m, n)
        jac_g = np.repeat(jac_g, baseline_hf.shape[-1], axis=-1)  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                baseline_hf * jac_g,  #  (nb_coef, m, n)
                g * baseline_jac_hf,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_chf(self, time, covar, *args, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)
        baseline_chf = self.baseline.chf(time, *args)  # (m, n)
        #  p == baseline.nb_params
        baseline_jac_chf = self.baseline.jac_chf(time, *args, asarray=True)  # (p, m, n)
        jac_g = np.repeat(jac_g, baseline_chf.shape[-1], axis=-1)  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                baseline_chf * jac_g,  #  (nb_coef, m, n)
                g * baseline_jac_chf,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac


class AcceleratedFailureTime(LifetimeRegression):
    # noinspection PyUnresolvedReferences
    r"""
    Accelerated failure time regression.

    The cumulative hazard function :math:`H` is linked to the multiplier
    function :math:`g` by the relation:

    .. math::

        H(t, x) = H_0\left(\dfrac{t}{g(\beta, x)}\right) = H_0(t e^{- \beta
        \cdot x})

    where :math:`x` is a vector of covariates, :math:`\beta` is the coefficient
    vector of the effect of covariates, :math:`H_0` is the baseline cumulative
    hazard function [1]_.

    |

    Parameters
    ----------
    baseline : FittableParametricLifetimeModel
        Any lifetime model that can be fitted.
    coefficients : tuple of floats (values can be None), default is (None,)
        Coefficients values of the covariate effects.

    Attributes
    ----------
    baseline : FittableParametricLifetimeModel
        The regression baseline model (lifetime model).
    covar_effect : CovarEffect
        The regression covariate effect.
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.). If the model is not fitted, the value is None.
    coefficients
    nb_params
    params
    params_names
    plot


    References
    ----------
    .. [1] Kalbfleisch, J. D., & Prentice, R. L. (2011). The statistical
        analysis of failure time data. John Wiley & Sons.

    See Also
    --------
    regression.ProportionalHazard : proportional hazard regression
    """

    def hf(self, time, covar, *args):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.hf(t0, *args) / self.covar_effect.g(covar)

    def chf(self, time, covar, *args):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.chf(t0, *args)

    def ichf(self, cumulative_hazard_rate, covar, *args):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape ``(nb_coef,)`` or ``(m, nb_coef)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given cumulative hazard rate(s).
        """
        return self.covar_effect.g(covar) * self.baseline.ichf(cumulative_hazard_rate, *args)

    def dhf(self, time, covar, *args):
        """
        The derivative of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.dhf(t0, *args) / self.covar_effect.g(covar) ** 2

    def jac_hf(self, time, covar, *args, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)
        t0 = time / g  # (m, n)
        # p == baseline.nb_params
        baseline_jac_hf_t0 = self.baseline.jac_hf(t0, *args, asarray=True)  # (p, m, n)
        baseline_hf_t0 = self.baseline.hf(t0, *args)  # (m, n)
        baseline_dhf_t0 = self.baseline.dhf(t0, *args)  # (m, n)
        jac_g = np.repeat(jac_g, baseline_hf_t0.shape[-1], axis=-1)  # (nb_coef, m, n)

        jac = np.concatenate(
            (
                -jac_g / g**2 * (baseline_hf_t0 + t0 * baseline_dhf_t0),  # (nb_coef, m, n) necessary to concatenate
                baseline_jac_hf_t0 / g,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac

    def jac_chf(self, time, covar, *args, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n_values,)`` or ``(n_assets, n_values)``.
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape (nb_coef,) or (m, nb_coef)
        *args : float or np.ndarray
            Additional arguments needed by the model.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """

        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = broadcast_time_covar_shapes(time.shape, covar.shape)  # (), (n,) or (m, n)
        time, covar = broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar, asarray=True)  # (nb_coef, m, 1)
        t0 = time / g  #  (m, n)
        # p == baseline.nb_params
        baseline_jac_chf_t0 = self.baseline.jac_chf(t0, *args, asarray=True)  # (p, m, n)
        baseline_hf_t0 = self.baseline.hf(t0, *args)  #  (m, n)
        jac_g = np.repeat(jac_g, baseline_hf_t0.shape[-1], axis=-1)  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                -jac_g / g * t0 * baseline_hf_t0,  #  (nb_coef, m, n)
                baseline_jac_chf_t0,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        if not asarray:
            return np.unstack(jac)
        return jac


class FrozenLifetimeRegression(FrozenParametricLifetimeModel):
    r"""
    Frozen lifetime regression.

    Parameters
    ----------
    regression : LifetimeRegression
        Any lifetime regression.
    covar : float or np.ndarray
        Covariate values to be frozen.
    *args : float or np.ndarray
        Additional arguments needed by the model to be frozen.

    Attributes
    ----------
    unfrozen_model : LifetimeRegression
        The unfrozen regression model.
    args : tuple of float or np.ndarray
        All the frozen arguments given and necessary to compute model functions.
    nb_assets : int
        Number of assets passed in frozen arguments. The data is mainly used to control numpy broadcasting and may not
        interest an user.

    Warnings
    --------
    This class is documented for the purpose of clarity and mainly address contributors or advance users. Actually, the
    recommanded way to instanciate a ``FrozenLifetimeRegression`` is use to ``freeze`` factory function.

    """

    def __init__(self, regression, covar, *args):
        super().__init__(regression, *(covar, *args))

    def unfreeze(self) -> LifetimeRegression:
        return self.unfrozen_model

    @property
    def nb_coef(self):
        """
        The number of coefficients

        Returns
        -------
        int
        """
        return self.unfrozen_model.nb_coef

    @property
    def covar(self):
        return self.args[0]

    @covar.setter
    def covar(self, value):
        self.args = (value,) + self.args[1:]

    def dhf(self, time):
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.dhf(time, self.args[0], *self.args[1:])

    def jac_hf(self, time, asarray=False):
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        return self.unfrozen_model.jac_hf(time, self.args[0], *self.args[1:], asarray=asarray)

    def jac_chf(self, time, asarray=False):
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        return self.unfrozen_model.jac_chf(time, self.args[0], *self.args[1:], asarray=asarray)

    def jac_sf(self, time, asarray=False):
        """
        The jacobian of the survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        return self.unfrozen_model.jac_sf(time, self.args[0], *self.args[1:], asarray=asarray)

    def jac_cdf(self, time, asarray=False):
        """
        The jacobian of the cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        return self.unfrozen_model.jac_cdf(time, self.args[0], *self.args[1:], asarray=asarray)

    def jac_pdf(self, time, asarray: bool = False):
        """
        The jacobian of the probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        asarray : bool, default is False

        Returns
        -------
        np.float64, np.ndarray or tuple of np.float64 or np.ndarray
            The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
            the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
            whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
            tuple when ``asarray`` is False.
        """
        return self.unfrozen_model.jac_pdf(time, self.args[0], *self.args[1:], asarray=asarray)



from abc import ABC, abstractmethod

import numpy as np
from scipy.optimize import Bounds, newton

from relife import FrozenParametricModel, ParametricModel
from relife.data import LifetimeData
from relife.likelihood import LikelihoodFromLifetimes
from relife.quadrature import (
    check_and_broadcast_bounds,
    legendre_quadrature,
    unweighted_laguerre_quadrature,
)

from ._plot import PlotParametricLifetimeModel


class ParametricLifetimeModel(ParametricModel, ABC):
    r"""Base class for lifetime model.

    This class defines the structure for creating lifetime model. It is a blueprint
    for implementing lifetime model expecting a variadic set of arguments.
    It expects implemantation of hazard functions (``hf``), cumulative hazard functions (``chf``),
    probability density function (``pdf``) and survival function (``sf``).
    Other functions are implemented by default but can be overridden by derived classes.

    Note:
        The abstract methods also provides a default implementation. One may not have to implement
        ``hf``, ``chf``, ``pdf`` and ``sf`` and just call ``super()`` to access the base implementation.

    Methods:
        hf: Abstract method to compute the hazard function.
        chf: Abstract method to compute the cumulative hazard function.
        sf: Abstract method to compute the survival function.
        pdf: Abstract method to compute the probability density function.

    Raises:
        NotImplementedError: Raised when an abstract method or feature in this
        class has not been implemented in a derived class.
    """

    @abstractmethod
    def sf(self, time, *args):
        if hasattr(self, "chf"):
            return np.exp(
                -self.chf(
                    time,
                    *args,
                )
            )
        elif hasattr(self, "pdf") and hasattr(self, "hf"):
            return self.pdf(time, *args) / self.hf(time, *args)
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                {class_name} must implement concrete sf function
                """
            )

    @abstractmethod
    def hf(self, time, *args):
        if hasattr(self, "pdf") and hasattr(self, "sf"):
            return self.pdf(time, *args) / self.sf(time, *args)
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                    {class_name} must implement concrete hf function
                    """
            )

    @abstractmethod
    def chf(self, time, *args):
        if hasattr(self, "sf"):
            return -np.log(self.sf(time, *args))
        elif hasattr(self, "pdf") and hasattr(self, "hf"):
            return -np.log(self.pdf(time, *args) / self.hf(time, *args))
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                {class_name} must implement concrete chf or at least concrete hf function
                """
            )

    @abstractmethod
    def pdf(self, time, *args):
        try:
            return self.sf(time, *args) * self.hf(time, *args)
        except NotImplementedError as err:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
            {class_name} must implement pdf or the above functions
            """
            ) from err

    def cdf(self, time, *args):
        return 1 - self.sf(time, *args)

    def ppf(self, probability, *args):
        return self.isf(1 - probability, *args)

    def median(self, *args):
        return self.ppf(0.5, *args)

    def isf(self, probability, *args):
        func = lambda x: self.sf(x, *args) - probability
        return newton(
            func,
            x0=np.zeros_like(probability),
            args=args,
        )

    def ichf(self, cumulative_hazard_rate, *args):
        func = lambda x: np.sum(self.chf(x, *args) - cumulative_hazard_rate)
        return newton(
            func,
            x0=np.zeros_like(cumulative_hazard_rate),
        )

    def rvs(self, size, *args, nb_assets=None, return_event=False, return_entry=False, seed=None):
        rs = np.random.RandomState(seed=seed)
        if nb_assets is not None:
            np_size = (nb_assets, size)
        else:
            np_size = size
        probability = rs.uniform(size=np_size)
        time = self.isf(probability, *args)
        event = np.ones_like(time, dtype=np.bool_) if isinstance(time, np.ndarray) else np.bool_(True)
        entry = np.zeros_like(time, dtype=np.float64) if isinstance(time, np.ndarray) else np.float64(0)
        if not return_event and not return_entry:
            return time
        elif return_event and not return_entry:
            return time, event
        elif not return_event and return_entry:
            return time, entry
        else:
            return time, event, entry

    @property
    def plot(self):
        """Provides access to plotting functionnalities"""
        return PlotParametricLifetimeModel(self)

    def ls_integrate(self, func, a, b, *args, deg=10):
        def integrand(x):
            #  x.shape == (deg,), (deg, n) or (deg, m, n), ie points of quadratures
            # fx : (d_1, ..., d_i, deg), (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
            x = np.asarray(x)
            fx = func(x)
            if fx.shape[-len(x.shape) :] != x.shape:
                raise ValueError(
                    f"""
                    func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
                    Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
                    """
                )
            if x.ndim == 3:  # reshape because model.pdf is tested only for input ndim <= 2
                xdeg, m, n = x.shape
                x = np.rollaxis(x, 1).reshape(m, -1)  # (m, deg*n), roll on m because axis 0 must align with m of args
                pdf = self.pdf(x, *args)  # (m, deg*n)
                pdf = np.rollaxis(pdf.reshape(m, xdeg, n), 1, 0)  #  (deg, m, n)
            else:  # ndim == 1 | 2
                # reshape to (1, deg*n) or (1, deg), ie place 1 on axis 0 to allow broadcasting with m of args
                pdf = self.pdf(x.reshape(1, -1), *args)  # (1, deg*n) or (1, deg)
                pdf = pdf.reshape(x.shape)  # (deg, n) or (deg,)

            # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
            return fx * pdf

        arr_a, arr_b = check_and_broadcast_bounds(a, b)  # (), (n,) or (m, n)
        if np.any(arr_a > arr_b):
            raise ValueError("Bound values a must be lower than values of b")

        bound_b = self.isf(1e-4, *args)  #  () or (m, 1), if (m, 1) then arr_b.shape == (m, 1) or (m, n)
        broadcasted_arrs = np.broadcast_arrays(arr_a, arr_b, bound_b)
        arr_a = broadcasted_arrs[0].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        arr_b = broadcasted_arrs[1].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        bound_b = broadcasted_arrs[2].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        is_inf = np.isinf(arr_b)  # () or (n,) or (m, n)
        arr_b = np.where(is_inf, bound_b, arr_b)
        integration = legendre_quadrature(
            integrand, arr_a, arr_b, deg=deg
        )  #  (d_1, ..., d_i), (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)
        is_inf, _ = np.broadcast_arrays(is_inf, integration)
        return np.where(
            is_inf,
            integration + unweighted_laguerre_quadrature(integrand, arr_b, deg=deg),
            integration,
        )

    def moment(self, n, *args):
        if n < 1:
            raise ValueError("order of the moment must be at least 1")
        func = lambda x: np.power(x, n)
        return self.ls_integrate(
            func,
            0.0,
            np.inf,
            *args,
            deg=100,
        )  #  high degree of polynome to ensure high precision

    def mean(self, *args):
        return self.moment(1, *args)

    def var(self, *args):
        return self.moment(2, *args) - self.moment(1, *args) ** 2

    def mrl(self, time, *args):
        sf = self.sf(time, *args)
        func = lambda x: np.asarray(x) - time
        ls = self.ls_integrate(func, time, np.inf, *args)
        if sf.ndim < 2:  # 2d to 1d or 0d
            ls = np.squeeze(ls)
        return ls / sf


class FittableParametricLifetimeModel(ParametricLifetimeModel, ABC):

    def __init__(self, **kwparams):
        super().__init__(**kwparams)
        self.fitting_results = None

    @abstractmethod
    def _get_initial_params(self, time, *args, event=None, entry=None, departure=None): ...

    @abstractmethod
    def _get_params_bounds(self):
        return Bounds(
            np.full(self.nb_params, np.finfo(float).resolution),
            np.full(self.nb_params, np.inf),
        )

    @abstractmethod
    def dhf(self, time, *args): ...

    @abstractmethod
    def jac_hf(self, time, *args, asarray=True): ...

    @abstractmethod
    def jac_chf(self, time, *args, asarray=True): ...

    @abstractmethod
    def jac_sf(self, time, *args, asarray=True): ...

    @abstractmethod
    def jac_pdf(self, time, *args, asarray=True): ...

    def fit(
        self,
        time,
        *args,
        event=None,
        entry=None,
        departure=None,
        **options,
    ):
        lifetime_data = LifetimeData(time, event=event, entry=entry, departure=departure, args=args)
        self.params = self._get_initial_params(time, *args, event=event, entry=entry, departure=departure)
        likelihood = LikelihoodFromLifetimes(self, lifetime_data)
        if "bounds" not in options:
            options["bounds"] = self._get_params_bounds()
        fitting_results = likelihood.maximum_likelihood_estimation(**options)
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self


class NonParametricLifetimeModel(ABC):

    @abstractmethod
    def fit(self, time, event=None, entry=None, departure=None): ...


class FrozenParametricLifetimeModel(FrozenParametricModel):

    def __init__(self, model, *args):
        super().__init__(model, *args)

    def hf(self, time):
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.hf(time, *self.args)

    def chf(self, time):
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.chf(time, *self.args)

    def sf(self, time):
        """
        The survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.sf(time, *self.args)

    def pdf(self, time):
        """
        The probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.pdf(time, *self.args)

    def mrl(self, time):
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.mrl(time, *self.args)

    def moment(self, n: int):
        """
        n-th order moment.

        Parameters
        ----------
        n : int
            Order of the moment (at least 1)

        Returns
        -------
        np.float64 or np.ndarray
        """
        return self.unfrozen_model.moment(n, *self.args)

    def mean(self):
        """
        The mean.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return self.unfrozen_model.moment(1, *self.args)

    def var(self):
        """
        The variance.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return self.unfrozen_model.moment(2, *self.args) - self.unfrozen_model.moment(1, *self.args) ** 2

    def isf(self, probability):
        """
        The inverse survival function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        return self.unfrozen_model.isf(probability, *self.args)

    def ichf(self, cumulative_hazard_rate):
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are (), (n,) or (m, n).

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        return self.unfrozen_model.ichf(cumulative_hazard_rate, *self.args)

    def cdf(self, time):
        """
        The cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.unfrozen_model.cdf(time, *self.args)

    def rvs(self, size, return_event=False, return_entry=False, seed=None):
        """
        Random variable sampling.

        Parameters
        ----------
        size : int, (int,) or (int, int)
            Size of the generated sample. If size is ``n`` or ``(n,)``, n samples are generated. If size is ``(m,n)``, a
            2d array of samples is generated.
        return_event : bool, default is False
            If True, returns event indicators along with the sample time values.
        return_entry : bool, default is False
            If True, returns corresponding entry values of the sample time values.
        seed : optional int, default is None
            Random seed used to fix random sampling.

        Returns
        -------
        float, ndarray or tuple of float or ndarray
            The sample values. If either ``return_event`` or ``return_entry`` is True, returns a tuple containing
            the time values followed by event values, entry values or both.
        """
        return self.unfrozen_model.rvs(
            size, *self.args, return_event=return_event, return_entry=return_entry, seed=seed
        )

    def ppf(self, probability):
        """
        The percent point function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given probability value(s).
        """
        return self.unfrozen_model.ppf(probability, *self.args)

    def median(self):
        """
        The median.

        Returns
        -------
        np.float64 or np.ndarray
        """
        return self.unfrozen_model.median(*self.args)

    def ls_integrate(self, func, a, b, deg=10):
        """
        Lebesgue-Stieltjes integration.

        Parameters
        ----------
        func : callable (in : 1 ndarray , out : 1 ndarray)
            The callable must have only one ndarray object as argument and one ndarray object as output
        a : ndarray (maximum number of dimension is 2)
            Lower bound(s) of integration.
        b : ndarray (maximum number of dimension is 2)
            Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
        deg : int, default 10
            Degree of the polynomials interpolation

        Returns
        -------
        np.ndarray
            Lebesgue-Stieltjes integral of func from `a` to `b`.
        """
        return self.unfrozen_model.ls_integrate(func, a, b, *self.args, deg=deg)



from typing import Literal, Optional, Self, overload

import numpy as np
from numpy.typing import NDArray

from relife.data import LifetimeData
from relife.lifetime_model._plot import (
    PlotECDF,
    PlotKaplanMeier,
    PlotNelsonAalen,
    PlotTurnbull,
)

from ._base import NonParametricLifetimeModel


class ECDF(NonParametricLifetimeModel):
    """
    Empirical Cumulative Distribution Function.
    """

    def __init__(self):
        self._sf: Optional[NDArray[np.void]] = None
        self._cdf: Optional[NDArray[np.void]] = None

    def fit(
        self,
        time: NDArray[np.float64],
        /,
        event: Optional[NDArray[np.float64]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray (1d or 2d)
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        departure : ndarray of float (1d), default is None
            Right truncations applied to lifetime values.
        """
        lifetime_data = LifetimeData(
            time,
            event=event,
            entry=entry,
            departure=departure,
        )
        timeline, counts = np.unique(lifetime_data.complete_or_right_censored.lifetime_values, return_counts=True)
        timeline = np.insert(timeline, 0, 0)

        dtype = np.dtype([("timeline", np.float64), ("estimation", np.float64), ("se", np.float64)])
        self._sf = np.empty((timeline.size,), dtype=dtype)
        self._cdf = np.empty((timeline.size,), dtype=dtype)

        self._sf["timeline"] = timeline
        self._cdf["timeline"] = timeline
        cdf = np.insert(np.cumsum(counts), 0, 0) / np.sum(counts)
        self._cdf["estimation"] = cdf
        self._sf["estimation"] = 1 - cdf
        se = np.sqrt(*(1 - cdf) / len(lifetime_data.complete_or_right_censored.lifetime_values))
        self._sf["se"] = se
        self._cdf["se"] = se
        return self

    @overload
    def sf(self, se: Literal[False] = False) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]: ...

    @overload
    def sf(
        self, se: Literal[True] = True
    ) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]: ...

    def sf(
        self, se: bool = False
    ) -> (
        Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]
        | Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]
    ):
        """
        The survival functions estimated values

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._sf is None:
            return None
        if se:
            return self._sf["timeline"], self._sf["estimation"], self._sf["se"]
        return self._sf["timeline"], self._sf["estimation"]

    @overload
    def cdf(self, se: Literal[False] = False) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]: ...

    @overload
    def cdf(
        self, se: Literal[True] = True
    ) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]: ...

    def cdf(
        self, se: bool = False
    ) -> (
        Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]
        | Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]
    ):
        """
        The cumulative distribution function estimated values

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._cdf is None:
            return None
        if se:
            return self._cdf["timeline"], self._cdf["estimation"], self._cdf["se"]
        return self._cdf["timeline"], self._cdf["estimation"]

    @property
    def plot(self) -> PlotECDF:
        return PlotECDF(self)


class KaplanMeier(NonParametricLifetimeModel):
    r"""Kaplan-Meier estimator.

    Compute the non-parametric Kaplan-Meier estimator (also known as the product
    limit estimator) of the survival function from lifetime data.

    Notes
    -----
    For a given time instant :math:`t` and :math:`n` total observations, this
    estimator is defined as:

    .. math::

        \hat{S}(t) = \prod_{i: t_i \leq t} \left( 1 - \frac{d_i}{n_i}\right)

    where :math:`d_i` is the number of failures until :math:`t_i` and
    :math:`n_i` is the number of assets at risk just prior to :math:`t_i`.

    The variance estimation is obtained by:

    .. math::

        \widehat{Var}[\hat{S}(t)] = \hat{S}(t)^2 \sum_{i: t_i \leq t}
        \frac{d_i}{n_i(n_i - d_i)}

    which is often referred to as Greenwood's formula.

    References
    ----------
    .. [1] Lawless, J. F. (2011). Statistical models and methods for lifetime
        data. John Wiley & Sons.

    .. [2] Kaplan, E. L., & Meier, P. (1958). Nonparametric estimation from
        incomplete observations. Journal of the American statistical
        association, 53(282), 457-481.

    """

    def __init__(self):
        self._sf: Optional[np.void] = None

    def fit(
        self,
        time: NDArray[np.float64],
        event: Optional[NDArray[np.float64]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray (1d or 2d)
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        departure : ndarray of float (1d), default is None
            Right truncations applied to lifetime values.
        inplace : boolean, default is True
            If true, estimations are stored in the object
        """

        lifetime_data = LifetimeData(
            time,
            event=event,
            entry=entry,
            departure=departure,
        )

        if lifetime_data.left_censoring is not None:
            raise ValueError("KaplanMeier does not take left censored lifetimes")
        timeline, unique_indices, counts = np.unique(
            lifetime_data.complete_or_right_censored.lifetime_values,
            return_inverse=True,
            return_counts=True,
        )
        death_set = np.zeros_like(timeline, int)  # death at each timeline step
        complete_observation_indic = np.zeros_like(
            lifetime_data.complete_or_right_censored.lifetime_values
        )  # just creating an array to fill it next line
        complete_observation_indic[lifetime_data.complete.lifetime_index] = 1
        np.add.at(death_set, unique_indices, complete_observation_indic)
        x_in = np.histogram(
            np.concatenate(
                (
                    lifetime_data.left_truncation.lifetime_values.flatten(),
                    np.array(
                        [
                            0
                            for _ in range(
                                len(lifetime_data.complete_or_right_censored.lifetime_values)
                                - len(lifetime_data.left_truncation.lifetime_values)
                            )
                        ]
                    ),  # TODO : remplacer ça par self.entry en définissant self.entry plus haut?
                )
            ),
            np.insert(timeline, 0, 0),
        )[0]
        x_out = np.insert(counts[:-1], 0, 0)
        at_risk_assets = np.cumsum(x_in - x_out)
        s = np.cumprod(1 - death_set / at_risk_assets)
        sf = np.insert(s, 0, 1)

        with np.errstate(divide="ignore"):
            var = s**2 * np.cumsum(
                np.where(
                    at_risk_assets > death_set,
                    death_set / (at_risk_assets * (at_risk_assets - death_set)),
                    0,
                )
            )
        se = np.sqrt(np.insert(var, 0, 0))
        timeline = np.insert(timeline, 0, 0)

        dtype = np.dtype([("timeline", np.float64), ("estimation", np.float64), ("se", np.float64)])
        self._sf = np.empty((timeline.size,), dtype=dtype)
        self._sf["timeline"] = timeline
        self._sf["estimation"] = sf
        self._sf["se"] = se
        return self

    @overload
    def sf(self, se: Literal[False] = False) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]: ...

    @overload
    def sf(
        self, se: Literal[True] = True
    ) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]: ...

    def sf(
        self, se: bool = False
    ) -> (
        Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]
        | Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]
    ):
        """
        The survival function estimation

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._sf is None:
            return None
        if se:
            return self._sf["timeline"], self._sf["estimation"], self._sf["se"]
        return self._sf["timeline"], self._sf["estimation"]

    @property
    def plot(self) -> PlotKaplanMeier:
        return PlotKaplanMeier(self)


class NelsonAalen(NonParametricLifetimeModel):
    r"""Nelson-Aalen estimator.

    Compute the non-parametric Nelson-Aalen estimator of the cumulative hazard
    function from lifetime data.

    Notes
    -----
    For a given time instant :math:`t` and :math:`n` total observations, this
    estimator is defined as:

    .. math::

        \hat{H}(t) = \sum_{i: t_i \leq t} \frac{d_i}{n_i}

    where :math:`d_i` is the number of failures until :math:`t_i` and
    :math:`n_i` is the number of assets at risk just prior to :math:`t_i`.

    The variance estimation is obtained by:

    .. math::

        \widehat{Var}[\hat{H}(t)] = \sum_{i: t_i \leq t} \frac{d_i}{n_i^2}

    Note that the alternative survivor function estimate:

    .. math::

        \tilde{S}(t) = \exp{(-\hat{H}(t))}

    is sometimes suggested for the continuous-time case.

    References
    ----------
    .. [1] Lawless, J. F. (2011). Statistical models and methods for lifetime
        data. John Wiley & Sons.
    """

    def __init__(self):
        self._chf: Optional[NDArray[np.void]] = None

    def fit(
        self,
        time: NDArray[np.float64],
        event: Optional[NDArray[np.float64]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray (1d or 2d)
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        departure : ndarray of float (1d), default is None
            Right truncations applied to lifetime values.
        """

        lifetime_data = LifetimeData(
            time,
            event=event,
            entry=entry,
            departure=departure,
        )

        if lifetime_data.left_censoring is not None:
            raise ValueError("NelsonAalen does not accept left censored or interval censored lifetimes")

        timeline, unique_indices, counts = np.unique(
            lifetime_data.complete_or_right_censored.lifetime_values,
            return_inverse=True,
            return_counts=True,
        )
        death_set = np.zeros_like(timeline, dtype=np.int64)  # death at each timeline step

        complete_observation_indic = np.zeros_like(
            lifetime_data.complete_or_right_censored.lifetime_values
        )  # just creating an array to fill it next line
        complete_observation_indic[lifetime_data.complete.lifetime_index] = 1

        np.add.at(death_set, unique_indices, complete_observation_indic)
        x_in = np.histogram(
            np.concatenate(
                (
                    lifetime_data.left_truncation.lifetime_values.flatten(),
                    np.array(
                        [
                            0
                            for _ in range(
                                len(lifetime_data.complete_or_right_censored.lifetime_values)
                                - len(lifetime_data.left_truncation.lifetime_values)
                            )
                        ]
                    ),  # TODO : remplacer ça par self.entry en définissant self.entry plus haut?
                )
            ),
            np.insert(timeline, 0, 0),
        )[0]
        x_out = np.insert(counts[:-1], 0, 0)
        at_risk_assets = np.cumsum(x_in - x_out)
        s = np.cumsum(death_set / at_risk_assets)
        var = np.cumsum(death_set / np.where(at_risk_assets == 0, 1, at_risk_assets**2))
        chf = np.insert(s, 0, 0)
        se = np.sqrt(np.insert(var, 0, 0))
        timeline = np.insert(timeline, 0, 0)

        dtype = np.dtype([("timeline", np.float64), ("estimation", np.float64), ("se", np.float64)])
        self._chf = np.empty((timeline.size,), dtype=dtype)
        self._chf["timeline"] = timeline
        self._chf["estimation"] = chf
        self._chf["se"] = se
        return self

    @overload
    def chf(self, se: Literal[False] = False) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]: ...

    @overload
    def chf(
        self, se: Literal[True] = True
    ) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]: ...

    def chf(
        self, se: bool = False
    ) -> (
        Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]
        | Optional[tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]]
    ):
        """
        The cumulative hazard function estimation

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._chf is None:
            return None
        if se:
            return self._chf["timeline"], self._chf["estimation"], self._chf["se"]
        return self._chf["timeline"], self._chf["estimation"]

    @property
    def plot(self) -> PlotNelsonAalen:
        return PlotNelsonAalen(self)


class Turnbull(NonParametricLifetimeModel):
    """Turnbull estimator."""

    def __init__(
        self,
        tol: Optional[float] = 1e-4,
        lowmem: Optional[bool] = False,
    ):
        super().__init__()
        self.tol = tol
        self.lowmem = lowmem
        self._sf: Optional[NDArray[np.void]] = None

    def fit(
        self,
        time: NDArray[np.float64],
        event: Optional[NDArray[np.float64]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
        inplace: bool = False,
    ) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray (1d or 2d)
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        departure : ndarray of float (1d), default is None
            Right truncations applied to lifetime values.
        inplace : boolean, default is True
            If true, estimations are stored in the object
        """

        lifetime_data = LifetimeData(
            time,
            event=event,
            entry=entry,
            departure=departure,
        )

        timeline_temp = np.unique(np.insert(lifetime_data.interval_censoring.lifetime_values.flatten(), 0, 0))
        timeline_len = len(timeline_temp)
        if not self.lowmem:
            event_occurence = (
                np.greater_equal.outer(
                    timeline_temp[:-1],
                    lifetime_data.interval_censoring.lifetime_values[
                        :, 0
                    ],  # or self.observed_lifetimes.interval_censored.values.T[0][i]
                )
                * np.less_equal.outer(
                    timeline_temp[1:],
                    lifetime_data.interval_censoring.lifetime_values[:, 1],
                )
            ).T

            s = self._estimate_with_high_memory(
                lifetime_data,
                timeline_len,
                event_occurence,
                timeline_temp,
            )

        else:
            len_censored_data = len(lifetime_data.interval_censoring.lifetime_values)
            event_occurence = []
            for i in range(len_censored_data):
                event_occurence.append(
                    np.where(
                        (lifetime_data.interval_censoring.lifetime_values[:, 0][i] <= timeline_temp[:-1])
                        & (timeline_temp[1:] <= lifetime_data.interval_censoring.lifetime_values[:, 1][i])
                    )[0][[0, -1]]
                )
            event_occurence = np.array(event_occurence)
            s = self._estimate_with_low_memory(
                lifetime_data,
                timeline_temp,
                timeline_len,
                event_occurence,
                len_censored_data,
            )

        ind_del = np.where(timeline_temp == np.inf)
        sf = np.delete(s, ind_del)
        timeline = np.delete(timeline_temp, ind_del)
        dtype = np.dtype([("timeline", np.float64), ("estimation", np.float64)])
        self._sf = np.empty((timeline.size,), dtype=dtype)
        self._sf["timeline"] = timeline
        self._sf["estimation"] = sf
        return self

    def sf(self) -> Optional[tuple[NDArray[np.float64], NDArray[np.float64]]]:
        """
        The survival function estimation

        Returns
        -------
        tuple of 2 ndarrays
            A tuple containing the timeline and the estimated values
        """
        if self._sf is None:
            return None
        return self._sf["timeline"], self._sf["estimation"]

    def _estimate_with_low_memory(
        self,
        lifetime_data: LifetimeData,
        timeline_temp,
        timeline_len,
        event_occurence,
        len_censored_data,
    ):

        d_tilde = np.histogram(
            np.searchsorted(timeline_temp, lifetime_data.complete.lifetime_values),
            bins=range(timeline_len + 1),
        )[0][1:]
        s = np.linspace(1, 0, timeline_len)
        res = 1
        count = 1
        while res > self.tol:
            p = -np.diff(s)  # écart entre les points de S (survival function) => proba of an event occuring at
            if np.sum(p == 0) > 0:
                p = np.where(
                    p == 0,
                    1e-5 / np.sum(p == 0),
                    p - 1e-5 / ((timeline_len - 1) - np.sum(p == 0)),
                )  # remplace 0 par 1e-5 (et enlève cette quantité des autres proba pr sommer à 1)

            x = [p[event_occurence[i, 0] : (event_occurence[i, 1] + 1)] for i in range(event_occurence.shape[0])]
            d = np.repeat(0, timeline_len - 1)
            for i in range(len_censored_data):
                d = np.add(
                    d,
                    np.append(
                        np.insert(x[i] / x[i].sum(), 0, np.repeat(0, event_occurence[i][0])),
                        np.repeat(0, timeline_len - event_occurence[i][1] - 2),
                    ),
                )
            d += d_tilde
            y = np.cumsum(d[::-1])[::-1]
            _unsorted_entry = lifetime_data.left_truncation.lifetime_values.flatten()

            y -= len(_unsorted_entry) - np.searchsorted(np.sort(_unsorted_entry), timeline_temp[1:])
            s_updated = np.array(np.cumprod(1 - d / y))
            s_updated = np.insert(s_updated, 0, 1)
            res = max(abs(s - s_updated))
            s = s_updated
            count += 1
        return s

    def _estimate_with_high_memory(
        self,
        lifetime_data: LifetimeData,
        timeline_len,
        event_occurence,
        timeline_temp,
    ):

        d_tilde = np.histogram(
            np.searchsorted(timeline_temp, lifetime_data.complete.lifetime_values),
            bins=range(timeline_len + 1),
        )[0][1:]
        s = np.linspace(1, 0, timeline_len)
        res = 1
        count = 1

        while res > self.tol:
            p = -np.diff(s)  # écart entre les points de S (survival function) => proba of an event occuring at
            if np.sum(p == 0) > 0:
                p = np.where(
                    p == 0,
                    1e-5 / np.sum(p == 0),
                    p - 1e-5 / ((timeline_len - 1) - np.sum(p == 0)),
                )  # remplace 0 par 1e-5 (et enlève cette quantité des autres proba pr sommer à 1)

            if np.any(event_occurence):
                event_occurence_proba = event_occurence * p.T
                d = (event_occurence_proba.T / event_occurence_proba.sum(axis=1)).T.sum(axis=0)
                d += d_tilde
            else:
                d = d_tilde
            y = np.cumsum(d[::-1])[::-1]
            _unsorted_entry = lifetime_data.left_truncation.lifetime_values.flatten()
            y -= len(_unsorted_entry) - np.searchsorted(np.sort(_unsorted_entry), timeline_temp[1:])
            s_updated = np.array(np.cumprod(1 - d / y))
            s_updated = np.insert(s_updated, 0, 1)
            res = max(abs(s - s_updated))
            s = s_updated
            count += 1
        return s

    @property
    def plot(self) -> PlotTurnbull:
        return PlotTurnbull(self)



from typing import Any, Optional, Self, Sequence

import numpy as np
from numpy.typing import NDArray

from relife._typing import _X, _Y, _Xs
from relife.lifetime_model import FittableParametricLifetimeModel
from relife.likelihood import FittingResults

from .base import FrozenStochasticProcess, StochasticProcess

class NonHomogeneousPoissonProcess(StochasticProcess[*_Xs]):
    lifetime_model: FittableParametricLifetimeModel[*_Xs]
    fitting_results: Optional[FittingResults]
    def __init__(self, lifetime_model: FittableParametricLifetimeModel[*_Xs]) -> None: ...
    def intensity(self, time: _X, *args: *_Xs) -> _Y: ...
    def cumulative_intensity(self, time: _X, *args: *_Xs) -> _Y: ...
    def freeze(self, *args: *_Xs): ...
    def sample(
        self,
        size: int,
        tf: float,
        *args: *_Xs,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ): ...
    def generate_failure_data(
        self,
        size: int,
        tf: float,
        *args: *_Xs,
        t0: float = 0.0,
        nb_assets: Optional[int] = None,
        seed: Optional[int] = None,
    ): ...
    def fit(
        self,
        ages_at_events: NDArray[np.float64],
        events_assets_ids: Sequence[str] | NDArray[np.int64],
        first_ages: Optional[NDArray[np.float64]] = None,
        last_ages: Optional[NDArray[np.float64]] = None,
        model_args: Optional[tuple[*_Xs]] = None,
        assets_ids: Optional[Sequence[str] | NDArray[np.int64]] = None,
        **options: Any,
    ) -> Self: ...

class FrozenNonHomogeneousPoissonProcess(FrozenStochasticProcess[*_Xs]):
    unfrozen_model: NonHomogeneousPoissonProcess[*_Xs]
    args: tuple[*_Xs]

    def __init__(self, model: NonHomogeneousPoissonProcess[*_Xs], *args: *_Xs) -> None: ...
    def intensity(self, time: _X) -> _Y: ...
    def cumulative_intensity(self, time: _X) -> _Y: ...
    def sample(
        self, size: int, tf: float, t0: float = 0.0, nb_assets: Optional[int] = None, seed: Optional[int] = None
    ): ...
    def generate_failure_data(
        self, size: int, tf: float, t0: float = 0.0, nb_assets: Optional[int] = None, seed: Optional[int] = None
    ): ...



from abc import ABC
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    Self,
    overload,
)

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import Bounds
from typing_extensions import override

from relife._typing import _B, _X, _Y, _Xs
from relife.lifetime_model import LifetimeRegression
from relife.likelihood import FittingResults

from ._base import FittableParametricLifetimeModel, ParametricLifetimeModel

class LifetimeDistribution(FittableParametricLifetimeModel[()], ABC):
    fitting_results: Optional[FittingResults]

    def sf(self, time: _X) -> _Y: ...
    def pdf(self, time: _X) -> _Y: ...
    @override
    def isf(self, probability: _X) -> _Y: ...
    @override
    def cdf(self, time: _X) -> _Y: ...
    @override
    def ppf(self, probability: _X) -> _Y: ...
    @override
    def median(self) -> np.float64: ...
    @override
    def moment(self, n: int) -> np.float64: ...
    @override
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        deg: int = 10,
    ) -> _Y: ...
    @overload
    def jac_sf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_sf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_sf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_cdf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_cdf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_cdf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_pdf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_pdf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_pdf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        *,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> _Y: ...
    @overload
    def rvs(
        self,
        size: int,
        *,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[False] = False,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B]: ...
    @overload
    def rvs(
        self,
        size: int,
        *,
        nb_assets: Optional[int] = None,
        return_event: Literal[False] = False,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _Y]: ...
    @overload
    def rvs(
        self,
        size: int,
        *,
        nb_assets: Optional[int] = None,
        return_event: Literal[True] = True,
        return_entry: Literal[True] = True,
        seed: Optional[int] = None,
    ) -> tuple[_Y, _B, _Y]: ...
    def rvs(
        self,
        size: int,
        *,
        nb_assets: Optional[int] = None,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> _Y | tuple[_Y, _Y] | tuple[_Y, _B] | tuple[_Y, _B, _Y]: ...
    def _get_initial_params(self, time, event=None, entry=None, departure=None) -> NDArray[np.float64]: ...
    def _get_params_bounds(self) -> Bounds: ...
    @override
    def fit(
        self,
        time: NDArray[np.float64],
        *,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
        **options: Any,
    ) -> Self: ...

class Exponential(LifetimeDistribution):
    def __init__(self, rate: Optional[float] = None) -> None: ...
    @property
    def rate(self) -> float: ...
    def hf(self, time: _X) -> _Y: ...
    def chf(self, time: _X) -> _Y: ...
    @override
    def mean(self) -> np.float64: ...
    @override
    def var(self) -> np.float64: ...
    @override
    def mrl(self, time: _X) -> _Y: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X) -> _Y: ...
    def dhf(self, time: _X) -> _Y: ...
    @overload
    def jac_hf(self, time: _X, *, asarray: Literal[False]) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(self, time: _X, *, asarray: Literal[True]) -> _Y: ...
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(self, time: _X, *, asarray: Literal[False]) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(self, time: _X, *, asarray: Literal[True]) -> _Y: ...
    def jac_chf(self, time: _X, *, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...

class Weibull(LifetimeDistribution):
    def __init__(self, shape: Optional[float] = None, rate: Optional[float] = None) -> None: ...
    @property
    def shape(self) -> float: ...
    @property
    def rate(self) -> float: ...
    def hf(self, time: _X) -> _Y: ...
    def chf(self, time: _X) -> _Y: ...
    @override
    def mean(self) -> np.float64: ...
    @override
    def var(self) -> np.float64: ...
    @override
    def mrl(self, time: _X) -> _Y: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X) -> _Y: ...
    def dhf(self, time: _X) -> _Y: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...

class Gompertz(LifetimeDistribution):
    def __init__(self, shape: float | None = None, rate: float | None = None) -> None: ...
    @property
    def shape(self) -> float: ...
    @property
    def rate(self) -> float: ...
    def hf(self, time: _X) -> _Y: ...
    def chf(self, time: _X) -> _Y: ...
    @override
    def mean(self) -> np.float64: ...
    @override
    def var(self) -> np.float64: ...
    @override
    def mrl(self, time: _X) -> _Y: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X) -> _Y: ...
    def dhf(self, time: _X) -> _Y: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...

class Gamma(LifetimeDistribution):
    def __init__(self, shape: float | None = None, rate: float | None = None) -> None: ...
    def _uppergamma(self, x: _X) -> _Y: ...
    def _jac_uppergamma_shape(self, x: _X) -> _Y: ...
    @property
    def shape(self) -> float: ...
    @property
    def rate(self) -> float: ...
    def hf(self, time: _X) -> _Y: ...
    def chf(self, time: _X) -> _Y: ...
    @override
    def mean(self) -> np.float64: ...
    @override
    def var(self) -> np.float64: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X) -> _Y: ...
    @override
    def mrl(self, time: _X) -> _Y: ...
    def dhf(self, time: _X) -> _Y: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_hf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: Literal[False],
    ) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: Literal[True],
    ) -> _Y: ...
    def jac_chf(
        self,
        time: _X,
        *,
        asarray: bool = True,
    ) -> tuple[_Y, ...] | _Y: ...

class LogLogistic(LifetimeDistribution):
    def __init__(self, shape: float | None = None, rate: float | None = None) -> None: ...
    @property
    def shape(self) -> float: ...
    @property
    def rate(self) -> float: ...
    def hf(self, time: _X) -> _Y: ...
    def chf(self, time: _X) -> _Y: ...
    @override
    def mean(self) -> np.float64: ...
    @override
    def var(self) -> np.float64: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X) -> _Y: ...
    @override
    def mrl(self, time: _X) -> _Y: ...
    def dhf(self, time: _X) -> _Y: ...
    @overload
    def jac_hf(self, time: _X, *, asarray: Literal[False]) -> tuple[_Y, ...]: ...
    @overload
    def jac_hf(self, time: _X, *, asarray: Literal[True]) -> _Y: ...
    def jac_hf(self, time: _X, *, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...
    @overload
    def jac_chf(self, time: _X, *, asarray: Literal[False]) -> tuple[_Y, ...]: ...
    @overload
    def jac_chf(self, time: _X, *, asarray: Literal[True]) -> _Y: ...
    def jac_chf(self, time: _X, *, asarray: bool = True) -> tuple[_Y, ...] | _Y: ...

class EquilibriumDistribution(ParametricLifetimeModel[*tuple[_X, *_Xs]]):
    baseline: ParametricLifetimeModel[*tuple[_X, *_Xs]]
    fitting_results = Optional[FittingResults]
    def __init__(self, baseline: ParametricLifetimeModel[*tuple[_X, *_Xs]]) -> None: ...
    @property
    def args_names(self) -> tuple[str, ...]: ...
    @override
    def cdf(self, time: _X, *args: *_Xs) -> _Y: ...
    def sf(self, time: _X, *args: *_Xs) -> _Y: ...
    def pdf(self, time: _X, *args: *_Xs) -> _Y: ...
    def hf(self, time: _X, *args: *_Xs) -> _Y: ...
    def chf(self, time: _X, *args: *_Xs) -> _Y: ...
    @override
    def isf(self, probability: _X, *args: *_Xs) -> _Y: ...
    @override
    def ichf(self, cumulative_hazard_rate: _X, *args: *_Xs) -> _Y: ...

class MinimumDistribution(ParametricLifetimeModel[*tuple[int | NDArray[np.int64], *_Xs]]):
    baseline: FittableParametricLifetimeModel[*_Xs]
    fitting_results = Optional[FittingResults]
    def __init__(self, baseline: LifetimeDistribution | LifetimeRegression) -> None: ...
    @override
    def sf(self, time: _X, n: int | NDArray[np.int64], *args: *_Xs) -> _Y: ...
    @override
    def pdf(self, time: _X, n: int | NDArray[np.int64], *args: *_Xs) -> _Y: ...
    @override
    def hf(self, time: _X, n: int | NDArray[np.int64], *args: *_Xs) -> _Y: ...
    @override
    def chf(self, time: _X, n: int | NDArray[np.int64], *args: *_Xs) -> _Y: ...
    @override
    def ichf(
        self,
        cumulative_hazard_rate: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
    ) -> _Y: ...
    def dhf(self, time: _X, n: int | NDArray[np.int64], *args: *_Xs) -> _Y: ...
    def jac_chf(
        self,
        time: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
        asarray: bool = False,
    ) -> _Y | tuple[_Y, ...]: ...
    def jac_hf(
        self,
        time: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
        asarray: bool = False,
    ) -> _Y | tuple[_Y, ...]: ...
    def jac_sf(
        self,
        time: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
        asarray: bool = False,
    ) -> _Y | tuple[_Y, ...]: ...
    def jac_cdf(
        self,
        time: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
        asarray: bool = False,
    ) -> _Y | tuple[_Y, ...]: ...
    def jac_pdf(
        self,
        time: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
        asarray: bool = False,
    ) -> _Y | tuple[_Y, ...]: ...
    @override
    def ls_integrate(
        self,
        func: Callable[[_X], _Y],
        a: _X,
        b: _X,
        n: int | NDArray[np.int64],
        *args: *_Xs,
        deg: int = 10,
    ) -> _Y: ...
    def fit(
        self,
        time: NDArray[np.float64],
        n: NDArray[np.int64],
        *args: NDArray[np.float64],
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        departure: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> Self: ...



from typing import Generic, TypeVarTuple

import numpy as np
from _typeshed import Incomplete
from matplotlib.axes import Axes
from numpy.typing import NDArray as NDArray

from relife.lifetime_model import (
    NonParametricLifetimeModel as NonParametricLifetimeModel,
)
from relife.lifetime_model import ParametricLifetimeModel as ParametricLifetimeModel

ALPHA_CI: float

def plot_prob_function(
    x: NDArray[np.float64],
    y: NDArray[np.float64],
    se: NDArray[np.float64] | None = None,
    ci_bounds: tuple[float, float] | None = None,
    label: str | None = None,
    drawstyle: str = "default",
    **kwargs,
) -> Axes: ...

Args = TypeVarTuple("Args")

class PlotParametricLifetimeModel(Generic[*Args]):
    model: Incomplete
    def __init__(self, model: ParametricLifetimeModel[*Args]) -> None: ...
    def sf(self, *args: *Args, **kwargs) -> Axes: ...
    def cdf(self, *args: *Args, **kwargs) -> Axes: ...
    def chf(self, *args: *Args, **kwargs) -> Axes: ...
    def hf(self, *args: *Args, **kwargs) -> Axes: ...
    def pdf(self, *args: *Args, **kwargs) -> Axes: ...

class PlotNonParametricLifetimeModel:
    model: Incomplete
    def __init__(self, model: NonParametricLifetimeModel) -> None: ...
    def plot(
        self, fname: str, plot_se: bool = True, ci_bounds=(0.0, 1.0), drawstyle: str = "steps-post", **kwargs
    ) -> Axes: ...

class PlotECDF(PlotNonParametricLifetimeModel):
    def sf(self, plot_se: bool = True, **kwargs) -> Axes: ...
    def cdf(self, plot_se: bool = True, **kwargs) -> Axes: ...

class PlotKaplanMeier(PlotNonParametricLifetimeModel):
    def sf(self, plot_se: bool = True, **kwargs) -> Axes: ...

class PlotNelsonAalen(PlotNonParametricLifetimeModel):
    def chf(self, plot_se: bool = True, **kwargs) -> Axes: ...

class PlotTurnbull(PlotNonParametricLifetimeModel):
    def sf(self, **kwargs) -> Axes: ...
